//------------------------------------------------------------------------------
// Inside the Metal Detector
// PI-2 Firmware
//------------------------------------------------------------------------------
// This project uses the PIC 12F1840
// Clock is run off the internal 32MHz oscillator
// Each instruction takes 4 clock cycles so time resolution is 0.125us
// Timer0 controls the main loop, has 8 bits plus prescaler.
// Timer1 controls delays, has 16 bits plus prescaler.

//--- Includes ---
#include <htc.h>
__CONFIG(FOSC_INTOSC & MCLRE_ON & PLLEN_ON & WDTE_ON & PWRTE_OFF & CP_OFF \
& BOREN_OFF & CLKOUTEN_OFF & IESO_OFF & FCMEN_OFF);

//--- Defines ---
#define INPUT_PIN    1
#define OUTPUT_PIN   0

#define TxPulse      LATA5
#define Sample       LATA1
#define DelayPot     RA4                // AN3
#define PWM          LATA2
#define ADC          RA0                // AN0, PI-3 & PI-4

#define TrisTxPulse  TRISA5
#define TrisSample   TRISA1
#define TrisDelayPot TRISA4
#define TrisPWM      TRISA2
#define TrisADC      TRISA0             // PI-3 & PI-4

#define Period        47                // 255-208 -> 1664us (~600 Hz)
#define TxPulseWidth  64735             // 65535-800 -> 100us
#define SampleWidth   65415             // 65535-120 -> 15us
#define Delay10us     65455             // 65535-80 -> 10us delay
#define AudioDelay    65335             // 65535-200 -> 25us delay
#define AudioWidth    63935             // 65535-1600 -> 200us

#define SelectSample() CHS2=0; CHS1=0; CHS0=0; ADFM=1;   // Select AN0 for signal capture (right just.)
#define SelectDelayPot() CHS2=0; CHS1=1; CHS0=1; ADFM=0; // Select AN3 for the pot (left just.)

//--- Variables ---
unsigned int Sample1Delay = 65535-120;  // 15us

//--- Prototypes ---
void Init(void);
void InitAdc(void);
void InitPwm(void);                     // PI-3 & PI-4
void InitTimer0(void);
void InitTimer1(void);
void Timer1Delay(unsigned int);
void ReadDelayPot(void);
void ReadSample(void);                  // PI-3 & PI-4
void ProcessSample(void);               // PI-3 & PI-4
void CreateAudio(void);                 // PI-3 & PI-4

//------------------------------------------------------------------------------
// Main -- enter the program here
//
void main(void)
{
  // First, set up the internal oscillator
  IRCF3=1; IRCF2=1; IRCF1=1; IRCF0=0;   // Internal osc. freq. = 32MHz
  WDTPS4=0; WDTPS3=0; WDTPS2=0;         // Set the WDT to 4ms
  WDTPS1=1; WDTPS0=0;
  CLRWDT();                             // Using the WDT prevents PI latch-up

  // Next, initialize everything
  Init();                               // below
  InitAdc();
  InitTimer1();                         // Do this BEFORE InitTimer0()
  InitTimer0();
  InitPwm();                            // PI-3 & PI-4

  // Now start the interrupt timer and idle
  TMR0IF = 0;                           // Clear Timer0 flag
  TMR0 = 200;                           // Set the timer
  while(1) {}                           // Keep the processor active
}

//------------------------------------------------------------------------------
// Initialization routines

// Initialize the I/O pins and set some flags
//
void Init(void)
{
  TrisTxPulse  = OUTPUT_PIN;            // Set TX pulse to output
  TrisSample   = OUTPUT_PIN;            // Set main sample to output
  TrisPWM      = OUTPUT_PIN;            // Set pwm to output
  TrisDelayPot = INPUT_PIN;             // Set delaypot to input
  TrisADC      = INPUT_PIN;             // Set ADC to input (PI-3 & PI-4)

  GIE = 1;                              // Master interrupt enable
  PEIE = 0;                             // Peripheral interrupt disable
  CCP1IE = 0;                           // Capture/compare interrupt disabled
  CM1CON0 = 0x07;                       // Turn off analog comparator

  TxPulse = 0;                          // Initialize starting pulse values
  Sample = 0;
}

// Initialize the ADC.
//
void InitAdc(void)
{
  ANSELA = 0;                           // Clear the entire ADC select register
  ANSA0  = 1;                           // Set RA0 (AN0) to be an ADC input
  ANSA4  = 1;                           // Set RA4 (AN3) to be an ADC input (PI-3)

  ADCON0 = 0;                           // Clear the entire ADC control register
                                        //   This sets the active ADC channel to A0
                                        //   which is what we want to start with.

  ADPREF1=0; ADPREF0=0;                 // Use VDD for Vref

  ADCS2=0; ADCS1=1; ADCS0=0;            // ADC clock = fosc/32
  ADON = 1;                             // Turn the ADC on
}

// Defined for PI-3 & PI-4
//
void InitPwm(void)
{
  PWM = 0;
}

// Timer0 is 8 bits with Div64, so delay is (255-n)*64*4/Fosc
// For Fosc = 32MHz Timer0 has 8us of resolution
//
void InitTimer0(void)
{
  TMR0IE = 1;                           // Enable Timer0 interrupt
  TMR0CS = 0;                           // Clock off Fosc/4
  PSA = 0;                              // Prescaler assigned to Timer0
  PS2=1; PS1=0; PS0=1;                  // Prescaler = /64 (8us resolution)
}

// Timer1 is 16 bits and clocked at 32MHz/4; use PS=1 for 0.125us resolution
//
void InitTimer1(void)
{
  TMR1CS1=0; TMR1CS0=0;                 // Use Fosc/4
  T1CKPS1=0; T1CKPS0=0;                 // Prescaler = /1 (125ns resolution)
  TMR1H=0; TMR1L=0;                     // Clear the registers
  TMR1GE = 0;                           // Gate enable is off
  TMR1IE = 0;                           // Don't allow an interrupt
  TMR1ON = 1;                           // Turn Timer1 on
}

//------------------------------------------------------------------------------
// Processing routines

// Delay (in us) = 65535-n
// Because of the fudge factor, ~5us is the minimum delay.
//
void Timer1Delay(unsigned int n)
{
  n += 36;                              // Adjust for fixed error (approx 4.5us)
  TMR1H = n >> 8;                       // Load the upper counter
  TMR1L = n;                            // Load the lower counter
  TMR1IF = 0;                           // Clear the flag
  while(TMR1IF == 0) {;}                // Wait for a flag
}

// ISR() is the interrupt service routine for Timer0. The main loop is 
// triggered off this interrupt and must be completed before the next 
// interrupt.
//
static void interrupt ISR(void)
{
  TMR0IF = 0;                           // Clear the Timer0 Int flag
  TMR0 = Period;                        // Reset the timer (+2 fudge factor)
  CLRWDT();                             // Using the WDT prevents PI latch-up

  TxPulse = 1;
  Timer1Delay(TxPulseWidth);            // Transmit pulse
  TxPulse = 0;

  Timer1Delay(Sample1Delay);            // TX-to-sample pulse delay
  Sample = 1;
  Timer1Delay(SampleWidth);             // Sample pulse 
  Sample = 0;

  SelectDelayPot();                     // Set the ADC to read the pot

  Timer1Delay(AudioDelay);              // Delay for audio
  PWM = 1;
  Timer1Delay(AudioWidth);              // Speaker pulse
  PWM = 0;

  // The remaining routines are where extra processing gets done. It is critical 
  // that all processing is complete before the next TMR0 interrupt occurs, which 
  // depends on the TX pulse rate, TX pulse width, and sampling time. If the pulse 
  // rate = 600Hz, pulse width = 100us, and max sampling time = 35us then the 
  // processing time available is 1667us - 100us - 35us = 1532us. With a 32MHz
  // clock we have an instruction cycle of 0.125us, so there is time for 12256 code
  // instructions, including calls and returns.

  ReadDelayPot();
}

// Read the pulse delay pot
//
void ReadDelayPot(void)
{
  static char adc;

  GO = 1;                               // Start the conversion
  while(GO == 1);                       // Wait for the ADC to finish

  adc = ADRESH >> 1;                    // Look at the 7 MSBs; range = 0 to 127
  Sample1Delay = 65535 - 120 - adc;     // 15us min, 31us max
}
