//------------------------------------------------------------------------------
// Inside the Metal Detector
// PI-4 Firmware
//------------------------------------------------------------------------------
// This project uses the PIC 12F1840
// Clock is run off the internal 32MHz oscillator
// Each instruction takes 4 clock cycles so time resolution is 0.125us
// Timer0 controls the main loop, has 8 bits plus prescaler.
// Timer1 controls delays, has 16 bits plus prescaler.

//--- Includes ---
#include <htc.h>
__CONFIG(FOSC_INTOSC & MCLRE_ON & PLLEN_ON & WDTE_ON & PWRTE_OFF & CP_OFF \
& BOREN_OFF & CLKOUTEN_OFF & IESO_OFF & FCMEN_OFF);

//--- Defines ---
#define INPUT_PIN    1
#define OUTPUT_PIN   0

#define TxPulse      LATA5
#define DelayPot     RA4                // AN3
#define PWM          RA2
#define ADC          RA0                // AN0

#define TrisTxPulse  TRISA5
#define TrisDelayPot TRISA4
#define TrisPWM      TRISA2
#define TrisADC      TRISA0

#define Period        47                // 255-208 -> 1664us (~600 Hz)
#define TxPulseWidth  64735             // 65535-800 -> 100us
#define Delay10us     65455             // 65535-80 -> 10us delay

#define SelectSample() CHS2=0; CHS1=0; CHS0=0; ADFM=1;   // Select AN0 for signal capture (right just.)
#define SelectDelayPot() CHS2=0; CHS1=1; CHS0=1; ADFM=0; // Select AN3 for the pot (left just.)

//--- Variables ---
unsigned int Sample1Delay = 65535-120;  // 15us
unsigned int CurrentSample;             // Holds the current preamp sample
unsigned int SampleAve = 0;             // Running average of 64 samples
unsigned int TrackAve = 0;              // Running average of 64*16 samples

//--- Prototypes ---
void Init(void);
void InitAdc(void);
void InitPwm(void);
void InitTimer0(void);
void InitTimer1(void);
void Timer1Delay(unsigned int);
void ReadDelayPot(void);
void ReadSample(void);
void ProcessSample(void);
void CreateAudio(void);

//------------------------------------------------------------------------------
// Main -- enter the program here
//
void main(void)
{
  // First, set up the internal oscillator
  IRCF3=1; IRCF2=1; IRCF1=1; IRCF0=0;   // Internal osc. freq. = 32MHz
  WDTPS4=0; WDTPS3=0; WDTPS2=0;         // Set the WDT to 4ms
  WDTPS1=1; WDTPS0=0;
  CLRWDT();                             // Using the WDT prevents PI latch-up

  // Next, initialize everything
  Init();                               // below
  InitAdc();
  InitTimer1();                         // Do this BEFORE InitTimer0()
  InitTimer0();
  InitPwm();

  // Now start the interrupt timer and idle
  TMR0IF = 0;                           // Clear Timer0 flag
  TMR0 = 200;                           // Set the timer
  while(1) {}                           // Keep the processor active
}

//------------------------------------------------------------------------------
// Initialization routines

// Initialize the I/O pins and set some flags
//
void Init(void)
{
  TrisTxPulse  = OUTPUT_PIN;            // Set TX pulse to output
  TrisPWM      = OUTPUT_PIN;            // Set pwm pin to output
  TrisDelayPot = INPUT_PIN;             // Set delaypot to input
  TrisADC      = INPUT_PIN;             // Set ADC to input

  GIE = 1;                              // Master interrupt enable
  PEIE = 0;                             // Peripheral interrupt disable
  CCP1IE = 0;                           // Capture/compare interrupt disabled
  CM1CON0 = 0x07;                       // Turn off analog comparator

  TxPulse = 0;                          // Initialize starting pulse values
}

// Initialize the ADC.
//
void InitAdc(void)
{
  ANSELA = 0;                           // Clear the entire ADC select register
  ANSA0 = 1;                            // Set RA0 to be an ADC input
  ANSA4 = 1;                            // Set RA4 to be an ADC input

  ADCON0 = 0;                           // Clear the entire ADC control register
                                        //   This sets the active ADC channel to A0
                                        //   which is what we want to start with.

  ADFVR1=1; ADFVR0=1;                   // Set the ADC internal ref to 4.096V
  ADPREF1=1; ADPREF0=1;                 // Use the ADC internal ref

  ADCS2=0; ADCS1=1; ADCS0=0;            // ADC clock = fosc/32
  ADON = 1;                             // Turn the ADC on
}

// Initialize the PWM. Frequency is (Fosc/4)/64/255 = 490Hz; 64 is the TMR2 prescaler 
// and 255 is the PR2 value.
//
void InitPwm(void)
{
  T2CON = 0;                            // Clear the Timer2 control register
  T2CKPS1=1; T2CKPS0=1;                 // Set TMR2 prescaler to 64
  PR2 = 0xFF;                           // Set PR2=255
  CCP1CON = 0b00001100;                 // Set the CCP register up for std PWM
  CCPR1L = 127;                         // Clear the CCPR1L register
  TMR2ON = 1;                           // Turn on TMR2
}

// Timer0 is 8 bits with Div64, so delay is (255-n)*64*4/Fosc
// For Fosc = 32MHz Timer0 has 8us of resolution
//
void InitTimer0(void)
{
  TMR0IE = 1;                           // Enable Timer0 interrupt
  TMR0CS = 0;                           // Clock off Fosc/4
  PSA = 0;                              // Prescaler assigned to Timer0
  PS2=1; PS1=0; PS0=1;                  // Prescaler = /64 (8us resolution)
}

// Timer1 is 16 bits and clocked at Fosc/4; use PS=1 for 0.125us resolution
//
void InitTimer1(void)
{
  TMR1CS1=0; TMR1CS0=0;                 // Use Fosc/4
  T1CKPS1=0; T1CKPS0=0;                 // Prescaler = /1 (125ns resolution)
  TMR1H=0; TMR1L=0;                     // Clear the registers
  TMR1GE = 0;                           // Gate enable is off
  TMR1IE = 0;                           // Don't allow an interrupt
  TMR1ON = 1;                           // Turn Timer1 on
}

//------------------------------------------------------------------------------
// Processing routines

// Delay (in us) = 65535-n
// Because of the fudge factor, 15us is the minimum delay.
//
void Timer1Delay(unsigned int n)
{
  n += 36;                              // Adjust for fixed error (approx 4.5us)
  TMR1H = n >> 8;                       // Load the upper counter
  TMR1L = n;                            // Load the lower counter
  TMR1IF = 0;                           // Clear the flag
  while(TMR1IF == 0) {;}                // Wait for a flag
}

// ISR() is the interrupt service routine for Timer0. The main loop is 
// triggered off this interrupt and must be completed before the next 
// interrupt.
//
static void interrupt ISR(void)
{
  TMR0IF = 0;                           // Clear the Timer0 Int flag
  TMR0 = Period;                        // Reset the timer (+2 fudge factor)
  CLRWDT();                             // Using the WDT prevents PI latch-up

  SelectSample();                       // See #defines

  TxPulse = 1;
  Timer1Delay(TxPulseWidth);            // Transmit pulse
  TxPulse = 0;

  Timer1Delay(Sample1Delay);            // TX-to-sample pulse delay

  // The remaining routines are where all the processing gets done. It is critical 
  // that all processing is complete before the next TMR0 interrupt occurs, which 
  // depends on the TX pulse rate, TX pulse width, and sampling time. If the pulse 
  // rate = 600Hz, pulse width = 100us, and max sampling time = 35us then the 
  // processing time available is 1667us - 100us - 35us = 1532us. With a 32MHz
  // clock we have an instruction cycle of 0.125us, so there is time for 12256 code
  // instructions, including calls and returns. Current processing code runs in 
  // about 415us.

  ReadSample();                         // Sample the signal
  SelectDelayPot();                     // See #defines
  ProcessSample();
  CreateAudio();
  ReadDelayPot();
}

// Read the pulse delay pot
//
void ReadDelayPot(void)
{
  static char adc;

  GO = 1;                               // Start the conversion
  while(GO == 1);                       // Wait for the ADC to finish

  adc = ADRESH >> 1;                    // Look at the 7 MSBs; range = 0 to 127
  Sample1Delay = 65535 - 120 - adc;     // 15us min, 31us max
}

// Sample the output of the preamp, store the result
//
void ReadSample(void)
{
  GO = 1;                               // Start the conversion
  while(GO == 1);                       // Wait for the ADC to finish

  CurrentSample = ADRESH & 0x03;        // Get the high byte
  CurrentSample <<= 8;                  // Move into place
  CurrentSample += ADRESL;              // Add the low byte
  CurrentSample = ~CurrentSample;       // Bit-wise invert
  CurrentSample &= 0x03FF;              // Mask off 6 MSBs
}

// Process the latest sample into the Sample & Tracking arrays
//
#define sampleAveShift 5
#define sampleArraySize 32
#define trackAveShift 6
#define trackArraySize 64

void ProcessSample(void)
{
  static unsigned int sampleArray[sampleArraySize];
  static unsigned int trackArray[trackArraySize];
  static char sampleIndex = 0, trackIndex = 0;
  static unsigned int tmp = 0;
  static char i, sampleCount;

  tmp += CurrentSample;                 // tmp is effectively a 12-bit value
  sampleCount++;

  if(sampleCount == 4)                  // If we've accumulated 4 samples, then store
  {
    sampleArray[sampleIndex] = tmp>>2;  // Store the current sample - 10 bits
    sampleCount = 0;
    tmp = 0;

    sampleIndex++;                      // Increment the array pointer
    if(sampleIndex == sampleArraySize)  // If we're wrapping around...
      sampleIndex = 0;                  // Set the pointer back to the beginning

    SampleAve = 0;
    for(i = 0; i < sampleArraySize; i++)// Average all the samples
      SampleAve += sampleArray[i];      // This is effectively a 15-bit value
  
    SampleAve >>= sampleAveShift;       // Divide back down to a 10-bit value

    trackArray[trackIndex] = SampleAve; // Store the average - 10 bits

    trackIndex++;                       // Increment the array pointer
    if(trackIndex == trackArraySize)    // If we're wrapping around...
      trackIndex = 0;                   // Set the pointer back to the beginning

    TrackAve = 0;
    for(i = 0; i < trackArraySize; i++) // Average all the averages
      TrackAve += trackArray[i];        // This is effectively a 16-bit value

    TrackAve >>= trackAveShift;         // Divide back down to a 10-bit value
  }
}

// Create a PWM based on the difference between SampleAve and TrackAve. THRESH is
// the threshold offset when SampleAve = TrackAve; when SampleAve is above TrackAve
// the PWM duty cycle is increased, which results in a louder volume.
//
#define THRESH 2

typedef union
{
  int value;
  unsigned char junk:6;
  unsigned char CCPR1L:8;
  unsigned char DC1B1:1;
  unsigned char DC1B0:1;
} DutyCycle;

void CreateAudio()
{
  static DutyCycle dc;

  dc.value = SampleAve - TrackAve + THRESH;
  if(dc.value < THRESH) dc.value = THRESH;
  if(dc.value > 127) dc.value = 127;

  PR2 = 255-dc.value;                   // VCO option

  CCPR1L = dc.CCPR1L;                   // Bits 2-9 here
  DC1B1 = dc.DC1B1;
  DC1B0 = dc.DC1B0;
}
